# pylint: disable=invalid-name,no-self-use
"""Declares :class:`OAuth2Ctrl`."""
import functools
import time
import uuid

from authlib.integrations.django_client import OAuth
from django.http import JsonResponse
from django.middleware.csrf import rotate_token
from django.urls import include
from django.urls import reverse
from django.urls import path

from iam.ext.idc.lib.oauth2.registry import OAuth2ProviderRegistry


class OAuth2Ctrl:
    """Controllers for all OAuth2 provider endpoints."""

    @property
    def urlpatterns(self):
        """Return the list of URL patterns exposed by :class:`OAuth2Ctrl`."""
        return ([
            path('login', self.login_using_default,
                name='login_default'),
            path('providers/default', self.get_default_provider,
                name='providers.default'),
            path('providers', self.list_providers,
                name='providers'),
            path('<str:using>/authorize', self.inject_provider(self.authorize),
                name='authorize'),
            path('<str:using>/login', self.inject_provider(self.login),
                name='login'),
        ], 'idc')

    @classmethod
    def as_namespace(cls, base_path, namespace='oauth2'):
        """Instantiate a :class:`OAuth2Ctrl` instance and return an iterable
        suitable for use with :func:`django.urls.include`.

        Args:
            base_path (str): the base path to mount the OAuth2 URL to. Must not
                include a leading slash.
            namespace (str): the URL namespace. Defaults to ``oauth2``.

        Returns:
            iterable
        """
        ctrl = cls(OAuth2ProviderRegistry(OAuth()), namespace=namespace)
        return path(base_path, include(ctrl.urlpatterns, namespace=namespace))

    def inject_provider(self, func):
        """Decorate `func` so that the provider is injected based on the
        `using` URL parameter.
        """
        @functools.wraps(func)
        def f(request, using, *args, **kwargs):
            try:
                return func(request, self.registry.get(using), *args, **kwargs)
            except KeyError:
                list_url = self.get_enabled_providers_uri(request)
                dto = {
                    'id': str(uuid.uuid4()),
                    'timestamp': int(time.time() * 1000),
                    'code': "IDENTITY_PROVIDER_DOES_NOT_EXIST",
                    'message': f"The selected identity provider '{using}' does not exist.",
                    'hint': f"Inspect {list_url} for a list of enabled identity providers.",
                    'audit': {
                        'message': "This error is recorded along with your IP address."
                    }
                }
                return JsonResponse(dto, status=404,
                    json_dumps_params={'indent': 2})

        return f

    def __init__(self, registry, namespace):
        self.registry = registry
        self.namespace = namespace

    def authorize(self, request, provider):
        """Invoked when the user is redirected by the SSO provider after the
        authentication dialog.
        """
        subject = provider.get_asserted_subject(request)
        if subject is not None:
            provider.login(request,subject)
        else:
            raise NotImplementedError
        return JsonResponse({'asid': subject.asid},
            json_dumps_params={'indent': 2})

    def get_authorize_uri(self, request, using): # pylint: disable=unused-argument
        """Return the absolute URL to which the :term:`Asserting Party` must
        redirect after the login dialog.
        """
        return reverse(f'{self.namespace}:authorize', kwargs={'using': using})

    def get_default_provider(self, request):
        """Return a JSON response containing the default OAuth2 provider."""
        provider = self.registry.get_default_provider()
        return JsonResponse({
            'description': provider.description,
            'loginUrl': self.get_login_uri(request, provider.name)
            }, json_dumps_params={'indent': 2})

    def get_enabled_providers_uri(self, request):
        """Return URL yielding the enabled providers."""
        return request.build_absolute_uri(reverse(f'{self.namespace}:providers'))

    def get_login_uri(self, request, using):
        """Return the login URL for the given provider."""
        return request.build_absolute_uri(
            reverse(f'{self.namespace}:login', kwargs={'using': using}))

    def list_providers(self, request):
        """Return a JSON response containing the enabled OAuth2 providers."""
        return JsonResponse({
            'providers': [{
            'description': x.description,
            'loginUrl': self.get_login_uri(request, x.name)
        } for x in self.registry.enabled()]}, json_dumps_params={'indent': 2})

    def login(self, request, provider):
        """Redirect the client to the OAuth2 provider login screen."""
        rotate_token(request)
        return provider.authorize_redirect(request,
            request.build_absolute_uri(self.get_authorize_uri(request, provider.name)))

    def login_using_default(self, request):
        """Login using the default provider."""
        return self.login(request, self.registry.get_default_provider())



"""Declares :class:`GoogleOIDCProvider`."""
from .oidc import OIDCGenericProvider


class GoogleOIDCProvider(OIDCGenericProvider):
    """Implements OAuth2 with Google using OpenID Connect (OIDC)."""
    metadata_url = "https://accounts.google.com/.well-known/openid-configuration"

    @staticmethod
    def get_client_kwargs(spec): # pylint: disable=arguments-differ
        """Return the client keyword arguments for the specified
        provider `spec`.
        """
        kwargs = {}
        if spec.get('scope'):
            kwargs['scope'] = str.join(' ', spec['scope'])
        return kwargs

    def get_asserted_subject_id(self, cs):
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        return f"{cs['sub']}@accounts.google.com"



"""Declares :class:`OAuth2ProviderRegistry`."""
from collections import defaultdict

from .....canon.idc import OAuth2ProviderSchema
from ..registry import BaseIdentityProviverRegistry
from .providers import GoogleOIDCProvider
from .providers import OIDCGenericProvider


class OAuth2ProviderRegistry(BaseIdentityProviverRegistry):
    """The registry for all OAuth2 providers."""
    provider_classes = defaultdict(lambda: OIDCGenericProvider, {
        'accounts.google.com': GoogleOIDCProvider
    })
    resource_kind = 'OAuth2Provider'
    schema_class = OAuth2ProviderSchema

    def __init__(self, oauth, *args, **kwargs):
        super().__init__()
        self.__oauth = oauth

    def get_provider_args(self, resource):
        """Returns the provider implementation class for the given
        resource.
        """
        return [self.__oauth], {}



"""Declares :class:`OIDCGenericProvider`."""
from .base import BaseOAuth2Provider


class OIDCGenericProvider(BaseOAuth2Provider):
    """Generic implementation of OAuth2 using the OpenID Connect (OIDC)
    protocol.
    """
    metadata_url = None

    @classmethod
    def fromresource(cls, resource, oauth, *args, **kwargs): # pylint: disable=arguments-differ
        """Create a new :class:`OIDCGenericProvider` from a provider
        resource.
        """
        meta = resource['metadata']
        spec = dict(resource['spec'])
        credentials = spec.pop('credentials')
        if not cls.metadata_url:
            raise NotImplementedError
        return cls.fromclient(
            oauth.register(
                name=meta['name'],
                server_metadata_url=cls.metadata_url,
                client_id=credentials['id'],
                client_secret=credentials['secret'],
                client_kwargs=cls.get_client_kwargs(spec)
            ), *args, **spec
        )

    @staticmethod
    def get_client_kwargs(*args, **kwargs): # pylint: disable=unused-argument
        """Return the client keyword arguments for the specified
        provider `spec`.
        """
        return None

    def get_asserted_subject_id(self, cs): # pylint: disable=invalid-name
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        raise NotImplementedError("Subclasses must override this method.")

    def is_default(self):
        """Return a boolean indicating if the provider is the default
        provider for a protocol.
        """
        return self.__is_default

    def is_enabled(self):
        """Return a boolean indicating if the provider is enabled
        for a protocol.
        """
        return True



# pylint: skip-file
import unittest
from collections import defaultdict

import marshmallow
import marshmallow.fields
import unimatrix.lib.test

from ..registry import BaseIdentityProviverRegistry


class default_provider_class:

    @classmethod
    def fromresource(cls, *args, **kwargs):
        return cls()


class typed_provider_class(default_provider_class):
    pass


@unimatrix.lib.test.unit
class BaseIdentityProviverRegistryTestCase(unittest.TestCase):

    class impl_class(BaseIdentityProviverRegistry):
        provider_classes = defaultdict(
            lambda: default_provider_class,
            {'test': typed_provider_class}
        )
        resource_kind = 'TestProvider'

        class schema_class(marshmallow.Schema):
            kind = marshmallow.fields.String(required=True)
            metadata = marshmallow.fields.Dict()

    def setUp(self):
        self.impl = self.impl_class()

    def test_get_provider_class_returns_default_without_type(self):
        cls = self.impl.get_provider_class({})
        self.assertEqual(cls, default_provider_class)

    def test_get_provider_class_returns_type(self):
        cls = self.impl.get_provider_class({'type': 'test'})
        self.assertEqual(cls, typed_provider_class)

    def test_add_updates_providers(self):
        providers = {}
        self.impl.add(providers, {
            'kind': self.impl_class.resource_kind,
            'metadata': {
                'name': 'foo'
            }
        })
        self.assertIn('foo', providers)

    def test_get_provider_args_defaults_to_empty(self):
        args, kwargs = self.impl.get_provider_args({})
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {})

    def get_provider_returns_instance(self):
        provider = self.impl.get_provider({
            'kind': self.impl_class.resource_kind,
            'metadata': {
                'name': 'foo'
            }
        })
        self.assertIsInstance(provider, default_provider_class)



# pylint: disable=invalid-name
"""Declares :class:`BaseOAuth2Provider`."""
import ioc

from ...provider import BaseProvider


class BaseOAuth2Provider(BaseProvider):
    """The base class for all OAuth2 providers."""

    @property
    def name(self):
        """Return the name with which the client was registered in the
        :mod:`authlib` library.
        """
        return self.__client.name

    @property
    def description(self):
        """Return the description as specified in the resource."""
        return self.__description

    def __init__(self, client, description, external, superusers, **kwargs):
        self.__client = client
        self.__description = description
        self.__external = external
        self.__superusers = superusers
        self.__is_default = kwargs.get('is_default')

    @classmethod
    def fromclient(cls, client, *args, **kwargs):
        """Return a new :class:`BaseOAuth2Provider` from a :mod:`authlib`
        OAuth2 client object.
        """
        return cls(client, *args, **kwargs)

    def authorize_redirect(self, request, redirect_to):
        """Initiate the OAuth2 login dialog by redirecting the user to the
        SSO provider login screen.
        """
        return self.__client.authorize_redirect(request, redirect_to)

    def get_asserted_subject(self, request):
        """Invoked after the OAuth2 provider redirects the client upon
        succesfully completing the authentication dialog.
        """
        return self.authenticate(
            request,
            token=self.__client.authorize_access_token(request),
            provider=self
        )

    def is_default(self):
        """Indicates if the provider is the default provider."""
        return self.__is_default

    @ioc.inject('subjects', 'iam.SubjectRepository')
    @ioc.inject('factory', 'iam.SubjectFactory')
    def token_to_subject(self, request, token, subjects, factory):
        """Exchange an access token for an OIDC ID token, and get or
        create a corresponding :term:`Subject` from the persistent
        data store.
        """
        cs = self.__client.parse_id_token(request, token)
        asid = self.get_asserted_subject_id(cs)
        try:
            sub = subjects.get(asid=asid)
        except subjects.DoesNotExist:
            sub = factory.new()
        sub.asid = asid
        sub.external = self.__external
        sub.is_superuser = self.is_superuser(cs)
        subjects.persist(sub)
        return sub

    def get_asserted_subject_id(self, cs):
        """Return the asserted subject identifier from the OAuth2 provider
        claims set.
        """
        raise NotImplementedError("Subclasses must override this method.")

    def is_superuser(self, cs):
        """Return a boolean indicating if the :term:`Subject` asserted by
        the claims set is a superuser.
        """
        yes = False
        for superuser in self.__superusers:
            claim, value = str.split(superuser, ':')
            if cs.get(claim) != value:
                continue
            yes = True
            break
        return yes



"""Declares :class:`IdentityConsumerConfig`."""
import ioc
from django.apps import AppConfig
from django.apps import apps

from .infra import SubjectRepository


class IdentityConsumerConfig(AppConfig):
    """Configures the :mod:`iam.ext.idc` package."""
    name = 'iam.ext.idc'
    label = 'idc'
    verbose_name = "IAM - Identity Consumer"

    def ready(self):
        """Invoked when the Django app registry has loaded all
        apps.
        """
        ioc.provide('iam.SubjectRepository', SubjectRepository())
        ioc.provide('iam.SubjectFactory', apps.get_model('idc.AssertedSubject'))



# pylint: disable=no-self-use,unused-argument,invalid-name
"""Declares :class:`BaseIdentityProviverRegistry`."""
import abc
import os
from collections import OrderedDict

import yaml


NOT_LOADED = object()


class BaseIdentityProviverRegistry(metaclass=abc.ABCMeta):
    """The base class for all identity provider registries."""
    provider_classes = abc.abstractproperty()
    resource_kind = abc.abstractproperty()
    schema_class = abc.abstractproperty()

    @property
    def secdir(self):
        """Returns the directory name holding the secrets (i.e. provider
        configurations) used by the :class:`BaseIdentityProviverRegistry`.
        """
        return os.path.join(self.__secdir, "iam.unimatrixone.io/providers")

    def __init__(self, *args, **kwargs):
        self.__providers = NOT_LOADED
        self.__schema = self.schema_class()
        self.__default = None
        self.__secdir = kwargs.get('secdir', os.getenv('APP_SECDIR'))

    def add(self, providers, resource):
        """Adds a new :term:`Identity Provider` to the registry
        dictionary `providers`.
        """
        providers[ resource['metadata']['name'] ] = self.get_provider(resource)
        return providers[ resource['metadata']['name'] ]

    def enabled(self):
        """Return a list containing all enabled providers."""
        if self.__providers == NOT_LOADED:
            self.load_providers()
        return list([x for x in self.__providers.values() if x.is_enabled()])

    def get_provider(self, resource):
        """Instantiate a new :class:`Provider` from the parameters
        specified in `resource`.
        """
        cls = self.get_provider_class(resource)
        args, kwargs = self.get_provider_args(resource)
        return cls.fromresource(resource, *args, **kwargs)

    def get_provider_args(self, resource):
        """Return the position and keyword arguments to instantiate
        a provider for this resource.
        """
        return [], {}

    def get_provider_class(self, resource):
        """Returns the provider implementation class for the given
        resource.
        """
        return self.provider_classes[ resource.get('type') ]

    def lookup(self, name):
        """Return the identity provider identified by `name`."""
        return self.__providers[name]

    def load_providers(self):
        """Load the providers of the specified :attr:`resource_kind` from
        the secrets directory.
        """
        providers = OrderedDict()
        for fn in os.listdir(self.secdir):
            fp = os.path.join(self.secdir, fn)
            for doc in yaml.load_all(open(fp), Loader=yaml.SafeLoader):
                if doc.get('kind') != self.resource_kind:
                    # Silently ignore unrecognized documents.
                    continue
                provider = self.add(providers, self.__schema.load(doc))
                if provider.is_default():
                    self.__default = provider

        self.__providers = providers

        # If there is no default provider at this point, select the first loaded
        # from the persistent configuration.
        if self.__default is None and bool(providers):
            self.__default = list(providers.values())[0]

    def get(self, name):
        """Return the identity provider identified by `name`. Load the providers
        from the persistent storage backend if not present.
        """
        if self.__providers == NOT_LOADED:
            self.load_providers()
        return self.lookup(name)

    def get_default_provider(self):
        """Return the default provider configured for this registry."""
        if self.__providers == NOT_LOADED:
            self.load_providers()
        return self.__default



# pylint: skip-file
import os
import tempfile
import unittest
from collections import defaultdict

import marshmallow
import marshmallow.fields
import unimatrix.lib.test

from ..registry import BaseIdentityProviverRegistry
from ..provider import BaseProvider


TEST_PROVIDERS = """
---
kind: MockProvider
metadata:
  name: test1
spec:
  default: false
---
kind: MockProvider
metadata:
  name: test2
spec:
  default: true
---
kind: MockProvider
metadata:
  name: test3
spec:
  default: true
  enabled: false
---
kind: OtherProvider
metadata:
  name: test4
"""


@unimatrix.lib.test.integration
class BaseIdentityProviverRegistryTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.secdir = tempfile.mkdtemp()
        os.makedirs(os.path.join(cls.secdir, 'iam.unimatrixone.io/providers'))

        cls.fp = os.path.join(cls.secdir,
            "iam.unimatrixone.io/providers", 'providers.yaml')
        with open(cls.fp, 'w') as f:
            f.write(TEST_PROVIDERS)

    def setUp(self):
        self.registry = MockRegistry(secdir=self.secdir)

    def test_load_from_disk(self):
        self.registry.load_providers()

    def test_enabled_shows_correct_number(self):
        self.assertEqual(len(self.registry.enabled()), 2)

    def test_get_provider_returns_named_provider(self):
        provider = self.registry.get('test1')

    def test_get_default_provider_returns_specified(self):
        provider = self.registry.get_default_provider()
        self.assertEqual(provider.resource['metadata']['name'], 'test3')

    def test_no_default_picks_first(self):
        fp = os.path.join(self.secdir,
            "iam.unimatrixone.io/providers", 'providers.yaml')
        with open(fp, 'w') as f:
            f.write(str.replace(TEST_PROVIDERS, 'default: true', 'default: false'))
        registry = MockRegistry(secdir=self.secdir)
        provider = registry.get_default_provider()
        self.assertEqual(provider.resource['metadata']['name'], 'test1')
    

class MockProvider(BaseProvider):

    @classmethod
    def fromresource(cls, resource):
        return cls(resource)

    def __init__(self, resource):
        self.resource = resource

    def is_default(self):
        return bool(self.resource['spec'].get('default'))

    def is_enabled(self):
        return bool(self.resource['spec'].get('enabled', True))


class MockRegistry(BaseIdentityProviverRegistry):
    provider_classes =  defaultdict(lambda: MockProvider)
    resource_kind = 'MockProvider'

    class schema_class(marshmallow.Schema):
        kind = marshmallow.fields.String(required=True)
        metadata = marshmallow.fields.Dict()
        spec = marshmallow.fields.Dict()
